# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any, cast
from xml.etree import ElementTree

from typing_extensions import override

from .compumethods.compumethod import CompuMethod
from .compumethods.createanycompumethod import create_any_compu_method_from_et
from .createanydiagcodedtype import create_any_diag_coded_type_from_et
from .decodestate import DecodeState
from .diagcodedtype import DiagCodedType
from .diagnostictroublecode import DiagnosticTroubleCode
from .dopbase import DopBase
from .encodestate import EncodeState
from .exceptions import DecodeError, EncodeError, odxassert, odxraise, odxrequire
from .linkeddtcdop import LinkedDtcDop
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .odxtypes import ParameterValue, odxstr_to_bool
from .physicaltype import PhysicalType
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class DtcDop(DopBase):
    """A DOP describing a diagnostic trouble code"""

    diag_coded_type: DiagCodedType
    physical_type: PhysicalType
    compu_method: CompuMethod
    dtcs_raw: list[DiagnosticTroubleCode | OdxLinkRef] = field(default_factory=list)
    linked_dtc_dops_raw: list[LinkedDtcDop] = field(default_factory=list)
    is_visible_raw: bool | None = None

    @property
    def dtcs(self) -> NamedItemList[DiagnosticTroubleCode]:
        return self._dtcs

    @property
    def linked_dtc_dops(self) -> NamedItemList[LinkedDtcDop]:
        return self._linked_dtc_dops

    @property
    def is_visible(self) -> bool:
        return self.is_visible_raw is True

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DtcDop":
        """Reads a DTC-DOP."""
        kwargs = dataclass_fields_asdict(DopBase.from_et(et_element, context))

        diag_coded_type = create_any_diag_coded_type_from_et(
            odxrequire(et_element.find("DIAG-CODED-TYPE")), context)
        physical_type = PhysicalType.from_et(odxrequire(et_element.find("PHYSICAL-TYPE")), context)
        compu_method = create_any_compu_method_from_et(
            odxrequire(et_element.find("COMPU-METHOD")),
            context,
            internal_type=diag_coded_type.base_data_type,
            physical_type=physical_type.base_data_type,
        )
        dtcs_raw: list[DiagnosticTroubleCode | OdxLinkRef] = []
        if (dtcs_elem := et_element.find("DTCS")) is not None:
            for dtc_proxy_elem in dtcs_elem:
                if dtc_proxy_elem.tag == "DTC":
                    dtcs_raw.append(DiagnosticTroubleCode.from_et(dtc_proxy_elem, context))
                elif dtc_proxy_elem.tag == "DTC-REF":
                    dtcs_raw.append(OdxLinkRef.from_et(dtc_proxy_elem, context))

        linked_dtc_dops_raw = [
            LinkedDtcDop.from_et(dtc_ref_elem, context)
            for dtc_ref_elem in et_element.iterfind("LINKED-DTC-DOPS/"
                                                    "LINKED-DTC-DOP")
        ]
        is_visible_raw = odxstr_to_bool(et_element.get("IS-VISIBLE"))

        return DtcDop(
            diag_coded_type=diag_coded_type,
            physical_type=physical_type,
            compu_method=compu_method,
            dtcs_raw=dtcs_raw,
            linked_dtc_dops_raw=linked_dtc_dops_raw,
            is_visible_raw=is_visible_raw,
            **kwargs)

    def __post_init__(self) -> None:
        self._init_finished = False

    @override
    def decode_from_pdu(self, decode_state: DecodeState) -> ParameterValue:

        int_trouble_code = self.diag_coded_type.decode_from_pdu(decode_state)

        if not self.compu_method.is_valid_internal_value(int_trouble_code):
            # TODO: How to prevent this?
            odxraise(
                f"DTC-DOP {self.short_name} could not convert the coded value "
                f" {repr(int_trouble_code)} to physical type {self.physical_type.base_data_type}.",
                DecodeError)
            return

        trouble_code = self.compu_method.convert_internal_to_physical(int_trouble_code)

        assert isinstance(trouble_code, int)

        dtcs = [x for x in self.dtcs if x.trouble_code == trouble_code]

        odxassert(len(dtcs) < 2, f"Multiple matching DTCs for trouble code 0x{trouble_code:06x}")

        if len(dtcs) == 1:
            # we found exactly one described DTC
            return dtcs[0]

        # the DTC was not specified. This probably means that the
        # diagnostic description file is incomplete.
        odxraise(
            f"Encountered DTC 0x{trouble_code:06x} which has not been defined "
            f"by the database", DecodeError)

        return DiagnosticTroubleCode(
            trouble_code=trouble_code,
            odx_id=cast(OdxLinkId, None),
            oid=None,
            short_name=f'DTC_{trouble_code:06x}',
            long_name=None,
            description=None,
            text=None,
            display_trouble_code=None,
            level=None,
            is_temporary_raw=None,
            sdgs=[],
        )

    def convert_to_numerical_trouble_code(self, dtc_value: ParameterValue) -> int:
        if isinstance(dtc_value, DiagnosticTroubleCode):
            return dtc_value.trouble_code
        elif isinstance(dtc_value, int):
            # assume that physical value is the trouble_code
            return dtc_value
        elif isinstance(dtc_value, str):
            # assume that physical value is the short_name
            dtcs = [dtc for dtc in self.dtcs if dtc.short_name == dtc_value]
            if len(dtcs) != 1:
                odxraise(f"No DTC named {dtc_value} found for DTC-DOP "
                         f"{self.short_name}.", EncodeError)
                return cast(int, None)

            return dtcs[0].trouble_code
        else:
            odxraise(
                f"The DTC-DOP {self.short_name} expected a"
                f" diagnostic trouble code but got {type(dtc_value).__name__}", EncodeError)
            return cast(int, None)

    @override
    def encode_into_pdu(self, physical_value: ParameterValue | None,
                        encode_state: EncodeState) -> None:
        if physical_value is None:
            odxraise(f"No DTC specified", EncodeError)
            return

        trouble_code = self.convert_to_numerical_trouble_code(physical_value)

        if not isinstance(trouble_code, int):
            odxraise()
        internal_trouble_code = self.compu_method.convert_physical_to_internal(trouble_code)
        if not isinstance(internal_trouble_code, int):
            odxraise()

        found = False
        for dtc in self.dtcs:
            if internal_trouble_code == dtc.trouble_code:
                found = True
                break

        if not found:
            odxraise(
                f"Unknown diagnostic trouble code {physical_value!r} "
                f"(0x{internal_trouble_code: 06x}) specified", EncodeError)

        self.diag_coded_type.encode_into_pdu(internal_trouble_code, encode_state)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        odxlinks = super()._build_odxlinks()

        odxlinks.update(self.compu_method._build_odxlinks())

        for dtc_proxy in self.dtcs_raw:
            if isinstance(dtc_proxy, DiagnosticTroubleCode):
                odxlinks.update(dtc_proxy._build_odxlinks())

        for linked_dtc_dop in self.linked_dtc_dops_raw:
            odxlinks.update(linked_dtc_dop._build_odxlinks())

        return odxlinks

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        self.compu_method._resolve_odxlinks(odxlinks)

        self._dtcs = NamedItemList[DiagnosticTroubleCode]()
        for dtc_proxy in self.dtcs_raw:
            if isinstance(dtc_proxy, DiagnosticTroubleCode):
                dtc_proxy._resolve_odxlinks(odxlinks)
                self._dtcs.append(dtc_proxy)
            elif isinstance(dtc_proxy, OdxLinkRef):
                dtc = odxlinks.resolve(dtc_proxy, DiagnosticTroubleCode)
                self._dtcs.append(dtc)

        for linked_dtc_dop in self.linked_dtc_dops_raw:
            linked_dtc_dop._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        # hack to avoid initializing the DtcDop object multiple
        # times. This is required, because the linked DTC DOP feature
        # requires the "parent" DTC DOPs to be fully initialized but
        # the standard does not define a formal ordering of DTC DOPs.
        if self._init_finished:
            return

        self._init_finished = True

        super()._resolve_snrefs(context)

        self.compu_method._resolve_snrefs(context)

        for dtc_proxy in self.dtcs_raw:
            if isinstance(dtc_proxy, DiagnosticTroubleCode):
                dtc_proxy._resolve_snrefs(context)

        for linked_dtc_dop in self.linked_dtc_dops_raw:
            linked_dtc_dop._resolve_snrefs(context)

        # add the inherited DTCs from linked DTC DOPs. Note that this
        # requires that there are no cycles in the "link-hierarchy"
        dtc_short_names = {dtc.short_name for dtc in self._dtcs}
        for linked_dtc_dop in self.linked_dtc_dops_raw:
            linked_dtc_dop.dtc_dop._resolve_snrefs(context)

            for dtc in linked_dtc_dop.dtc_dop.dtcs:
                if dtc.short_name in dtc_short_names:
                    # we already have a DTC with that name. Since we
                    # are not supposed to overwrite the local DTCs, we
                    # skip processing this one. TODO: Are inheritance
                    # conflicts for DTCs allowed?
                    continue

                if dtc.short_name in linked_dtc_dop.not_inherited_dtc_snrefs:
                    # DTC is explicitly not inherited
                    continue

                self._dtcs.append(dtc)
                dtc_short_names.add(dtc.short_name)

        # at this place, the linked DTC DOPs exhibit .short_name, so
        # we can create a NamedItemList...
        self._linked_dtc_dops = NamedItemList(self.linked_dtc_dops_raw)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .element import IdentifiableElement
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .odxtypes import odxstr_to_bool
from .snrefcontext import SnRefContext
from .specialdatagroup import SpecialDataGroup
from .text import Text
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class DiagnosticTroubleCode(IdentifiableElement):
    trouble_code: int
    display_trouble_code: str | None = None
    text: Text
    level: int | None = None
    sdgs: list[SpecialDataGroup] = field(default_factory=list)

    is_temporary_raw: bool | None = None

    @property
    def is_temporary(self) -> bool:
        return self.is_temporary_raw is True

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DiagnosticTroubleCode":
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        trouble_code = int(odxrequire(et_element.findtext("TROUBLE-CODE")))
        display_trouble_code = et_element.findtext("DISPLAY-TROUBLE-CODE")
        text = Text.from_et(odxrequire(et_element.find("TEXT")), context)
        level = None
        if (level_str := et_element.findtext("LEVEL")) is not None:
            level = int(level_str)
        sdgs = [SpecialDataGroup.from_et(sdge, context) for sdge in et_element.iterfind("SDGS/SDG")]

        is_temporary_raw = odxstr_to_bool(et_element.attrib.get("IS-TEMPORARY"))

        return DiagnosticTroubleCode(
            trouble_code=trouble_code,
            text=text,
            display_trouble_code=display_trouble_code,
            level=level,
            sdgs=sdgs,
            is_temporary_raw=is_temporary_raw,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result: dict[OdxLinkId, Any] = {}

        result[self.odx_id] = self

        for sdg in self.sdgs:
            result.update(sdg._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        for sdg in self.sdgs:
            sdg._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        for sdg in self.sdgs:
            sdg._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .diagcomm import DiagComm
from .inputparam import InputParam
from .nameditemlist import NamedItemList
from .negoutputparam import NegOutputParam
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .outputparam import OutputParam
from .progcode import ProgCode
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class SingleEcuJob(DiagComm):
    """A single ECU job is a diagnostic communication primitive.

    A single ECU job is more complex than a diagnostic service and is
    not provided natively by the ECU.  In particular, the job is
    defined in external programs which are referenced by the attribute
    `.prog_codes`.

    In contrast to "multiple ECU jobs", a single ECU job only involves
    calls to services provided by a single ECU.

    Single ECU jobs are defined in section 7.3.5.7 of the ASAM MCD-2
    standard.
    """

    prog_codes: list[ProgCode] = field(default_factory=list)
    input_params: NamedItemList[InputParam] = field(default_factory=NamedItemList)
    output_params: NamedItemList[OutputParam] = field(default_factory=NamedItemList)
    neg_output_params: NamedItemList[NegOutputParam] = field(default_factory=NamedItemList)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "SingleEcuJob":
        kwargs = dataclass_fields_asdict(DiagComm.from_et(et_element, context))

        prog_codes = [
            ProgCode.from_et(pc_elem, context)
            for pc_elem in et_element.iterfind("PROG-CODES/PROG-CODE")
        ]

        input_params = NamedItemList([
            InputParam.from_et(el, context)
            for el in et_element.iterfind("INPUT-PARAMS/INPUT-PARAM")
        ])
        output_params = NamedItemList([
            OutputParam.from_et(el, context)
            for el in et_element.iterfind("OUTPUT-PARAMS/OUTPUT-PARAM")
        ])
        neg_output_params = NamedItemList([
            NegOutputParam.from_et(el, context)
            for el in et_element.iterfind("NEG-OUTPUT-PARAMS/NEG-OUTPUT-PARAM")
        ])

        return SingleEcuJob(
            prog_codes=prog_codes,
            input_params=input_params,
            output_params=output_params,
            neg_output_params=neg_output_params,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = super()._build_odxlinks()

        for prog_code in self.prog_codes:
            result.update(prog_code._build_odxlinks())
        for input_param in self.input_params:
            result.update(input_param._build_odxlinks())
        for output_param in self.output_params:
            result.update(output_param._build_odxlinks())
        for neg_output_param in self.neg_output_params:
            result.update(neg_output_param._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        for prog_code in self.prog_codes:
            prog_code._resolve_odxlinks(odxlinks)
        for input_param in self.input_params:
            input_param._resolve_odxlinks(odxlinks)
        for output_param in self.output_params:
            output_param._resolve_odxlinks(odxlinks)
        for neg_output_param in self.neg_output_params:
            neg_output_param._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        context.single_ecu_job = self

        super()._resolve_snrefs(context)

        for prog_code in self.prog_codes:
            prog_code._resolve_snrefs(context)
        for input_param in self.input_params:
            input_param._resolve_snrefs(context)
        for output_param in self.output_params:
            output_param._resolve_snrefs(context)
        for neg_output_param in self.neg_output_params:
            neg_output_param._resolve_snrefs(context)

        context.single_ecu_job = None



# SPDX-License-Identifier: MIT
from enum import Enum


class DiagClassType(Enum):
    STARTCOMM = "STARTCOMM"
    STOPCOMM = "STOPCOMM"
    VARIANTIDENTIFICATION = "VARIANTIDENTIFICATION"
    READ_DYN_DEFINED_MESSAGE = "READ-DYN-DEFINED-MESSAGE"
    DYN_DEF_MESSAGE = "DYN-DEF-MESSAGE"
    CLEAR_DYN_DEF_MESSAGE = "CLEAR-DYN-DEF-MESSAGE"



# SPDX-License-Identifier: MIT
import re
import textwrap

import markdownify
from rich import print as rich_print
from rich.padding import Padding as RichPadding
from rich.table import Table as RichTable

from ..description import Description
from ..diaglayers.diaglayer import DiagLayer
from ..diagservice import DiagService
from ..parameters.codedconstparameter import CodedConstParameter
from ..parameters.nrcconstparameter import NrcConstParameter
from ..parameters.parameter import Parameter
from ..parameters.physicalconstantparameter import PhysicalConstantParameter
from ..parameters.systemparameter import SystemParameter
from ..parameters.valueparameter import ValueParameter
from ..singleecujob import SingleEcuJob


def format_desc(description: Description, indent: int = 0) -> str:
    # Collapse whitespaces
    desc = re.sub(r"\s+", " ", str(description))
    # Covert XHTML to Markdown
    desc = markdownify.markdownify(desc)
    # Collapse blank lines
    desc = re.sub(r"(\n\s*)+\n+", "\n", desc).strip()
    # add indentation
    desc = textwrap.indent(desc, " " * indent)

    return desc


def print_diagnostic_service(service: DiagService,
                             print_params: bool = False,
                             print_pre_condition_states: bool = False,
                             print_state_transitions: bool = False,
                             print_audiences: bool = False,
                             allow_unknown_bit_lengths: bool = False) -> None:

    rich_print(f" Service '{service.short_name}':")

    if service.description:
        desc = format_desc(service.description, indent=3)
        rich_print(f"  Description: " + desc)

    if print_pre_condition_states and len(service.pre_condition_states) > 0:
        pre_condition_states_short_names = [
            pre_condition_state.short_name for pre_condition_state in service.pre_condition_states
        ]
        rich_print(f"  Pre-Condition States: {', '.join(pre_condition_states_short_names)}")

    if print_state_transitions and len(service.state_transitions) > 0:
        state_transitions = [
            f"{state_transition.source_snref} -> {state_transition.target_snref}"
            for state_transition in service.state_transitions
        ]
        rich_print(f"  State Transitions: {', '.join(state_transitions)}")

    if print_audiences and service.audience:
        enabled_audiences_short_names = [
            enabled_audience.short_name for enabled_audience in service.audience.enabled_audiences
        ]
        rich_print(f"  Enabled Audiences: {', '.join(enabled_audiences_short_names)}")

    if print_params:
        print_service_parameters(service, allow_unknown_bit_lengths=allow_unknown_bit_lengths)


def print_service_parameters(service: DiagService,
                             *,
                             allow_unknown_bit_lengths: bool = False) -> None:
    # prints parameter details of request, positive response and
    # negative response of diagnostic service

    # Request
    if service.request:
        rich_print(f"  Request '{service.request.short_name}':")
        const_prefix = service.request.coded_const_prefix()
        rich_print(
            f"    Identifying Prefix: 0x{const_prefix.hex().upper()} ({bytes(const_prefix)!r})")
        rich_print(f"    Parameters:")
        param_table = extract_parameter_tabulation_data(service.request.parameters)
        rich_print(RichPadding(param_table, pad=(0, 0, 0, 4)))
        rich_print()
    else:
        rich_print(f"  No Request!")

    # Positive Responses
    if not service.positive_responses:
        rich_print(f"  No positive responses")

    for resp in service.positive_responses:
        rich_print(f"  Positive Response '{resp.short_name}':")
        rich_print(f"   Parameters:\n")
        table = extract_parameter_tabulation_data(list(resp.parameters))
        rich_print(RichPadding(table, pad=(0, 0, 0, 4)))
        rich_print()

    # Negative Response
    if not service.negative_responses:
        rich_print(f"  No negative responses")

    for resp in service.negative_responses:
        rich_print(f" Negative Response '{resp.short_name}':")
        rich_print(f"   Parameters:\n")
        table = extract_parameter_tabulation_data(list(resp.parameters))
        rich_print(RichPadding(table, pad=(0, 0, 0, 4)))
        rich_print()

    rich_print("\n")


def extract_service_tabulation_data(services: list[DiagService],
                                    *,
                                    additional_columns: list[tuple[str, list[str]]] | None = None
                                   ) -> RichTable:
    """Extracts data of diagnostic services into Dictionary which can
    be printed by tabulate module
    """

    # Create Rich table
    table = RichTable(
        title="", show_header=True, header_style="bold cyan", border_style="blue", show_lines=True)

    name_column: list[str] = []
    semantic_column: list[str] = []
    request_column: list[str] = []

    for service in services:
        name_column.append(service.short_name)
        semantic_column.append(service.semantic or "")

        if service.request:
            prefix = service.request.coded_const_prefix()
            request_column.append(f"0x{str(prefix.hex().upper())[:32]}...") if len(
                prefix) > 32 else request_column.append(f"0x{str(prefix.hex().upper())}")
        else:
            request_column.append("")

    table.add_column("Name", style="green")
    table.add_column("Semantic", justify="left", style="white")
    table.add_column("Request", justify="left", style="white")
    if additional_columns is not None:
        for ac_title, _ in additional_columns:
            table.add_column(ac_title, justify="left", style="white")

        rows = zip(
            name_column,
            semantic_column,
            request_column,
            *[ac[1] for ac in additional_columns],
            strict=False)
        for row in rows:
            table.add_row(*map(str, row))

    return table


def extract_parameter_tabulation_data(parameters: list[Parameter]) -> RichTable:
    # extracts data of parameters of diagnostic services into
    # a RichTable object that can be printed

    # Create Rich table
    table = RichTable(
        title="", show_header=True, header_style="bold cyan", border_style="blue", show_lines=True)

    # Add columns with appropriate styling
    table.add_column("Name", style="green")
    table.add_column("Byte Position", justify="right", style="yellow")
    table.add_column("Bit Length", justify="right", style="yellow")
    table.add_column("Semantic", justify="left", style="white")
    table.add_column("Parameter Type", justify="left", style="white")
    table.add_column("Data Type", justify="left", style="white")
    table.add_column("Value", justify="left", style="yellow")
    table.add_column("Value Type", justify="left", style="white")
    table.add_column("Linked DOP", justify="left", style="white")

    name_column: list[str] = []
    byte_column: list[str] = []
    bit_length_column: list[str] = []
    semantic_column: list[str] = []
    param_type_column: list[str] = []
    value_column: list[str] = []
    value_type_column: list[str] = []
    data_type_column: list[str] = []
    dop_column: list[str] = []

    for param in parameters:
        name_column.append(param.short_name)
        byte_column.append("" if param.byte_position is None else str(param.byte_position))
        semantic_column.append(param.semantic or "")
        param_type_column.append(param.parameter_type)
        length = 0
        if param.get_static_bit_length() is not None:
            n = param.get_static_bit_length()
            bit_length_column.append("" if n is None else str(n))
            length = (n or 0) // 4
        else:
            bit_length_column.append("")
        if isinstance(param, CodedConstParameter):
            if isinstance(param.coded_value, int):
                value_column.append(f"0x{param.coded_value:0{length}X}")
            elif isinstance(param.coded_value, bytes) or isinstance(param.coded_value, bytearray):
                value_column.append(f"0x{param.coded_value.hex().upper()}")
            else:
                value_column.append(f"{param.coded_value!r}")
            data_type_column.append(param.diag_coded_type.base_data_type.name)
            value_type_column.append('coded value')
            dop_column.append("")
        elif isinstance(param, NrcConstParameter):
            data_type_column.append(param.diag_coded_type.base_data_type.name)
            value_column.append(str(param.coded_values))
            value_type_column.append('coded values')
            dop_column.append("")
        elif isinstance(param, (PhysicalConstantParameter, SystemParameter, ValueParameter)):
            # this is a hack to make this routine work for parameters
            # which reference DOPs of a type that a is not yet
            # internalized. (all parameter objects of the tested types
            # are supposed to have a DOP.)
            param_dop = getattr(param, "_dop", None)

            if param_dop is not None:
                dop_column.append(param_dop.short_name)

            if param_dop is not None and (phys_type := getattr(param, "physical_type",
                                                               None)) is not None:
                data_type_column.append(phys_type.base_data_type.name)
            else:
                data_type_column.append("")
            if isinstance(param, PhysicalConstantParameter):
                if isinstance(param.physical_constant_value, bytes) or isinstance(
                        param.physical_constant_value, bytearray):
                    value_column.append(f"0x{param.physical_constant_value.hex().upper()}")
                else:
                    value_column.append(f"{param.physical_constant_value!r}")
                value_type_column.append('constant value')
            elif isinstance(param, ValueParameter) and param.physical_default_value is not None:
                if isinstance(param.physical_default_value, bytes) or isinstance(
                        param.physical_default_value, bytearray):
                    value_column.append(f"0x{param.physical_default_value.hex().upper()}")
                else:
                    value_column.append(f"{param.physical_default_value!r}")
                value_type_column.append('default value')
            else:
                value_column.append("")
                value_type_column.append("")
        else:
            value_column.append("")
            data_type_column.append("")
            value_type_column.append("")
            dop_column.append("")

    # Add all rows at once by zipping dictionary values
    rows = zip(
        name_column,
        byte_column,
        bit_length_column,
        semantic_column,
        param_type_column,
        data_type_column,
        value_column,
        value_type_column,
        dop_column,
        strict=False)
    for row in rows:
        table.add_row(*map(str, row))

    return table


def print_dl_metrics(variants: list[DiagLayer]) -> None:
    """
    Print diagnostic layer metrics using Rich tables.
    Args:
        variants: List of diagnostic layer variants to analyze
    """
    # Create Rich table
    table = RichTable(
        title="", show_header=True, header_style="bold cyan", border_style="blue", show_lines=True)

    # Add columns with appropriate styling
    table.add_column("Name", style="green")
    table.add_column("Variant Type", style="magenta")
    table.add_column("Number of Services", justify="right", style="yellow")
    table.add_column("Number of DOPs", justify="right", style="yellow")
    table.add_column("Number of communication parameters", justify="right", style="yellow")

    # Process each variant
    for variant in variants:
        assert isinstance(variant, DiagLayer)
        all_services: list[DiagService | SingleEcuJob] = sorted(
            variant.services, key=lambda x: x.short_name)
        ddds = variant.diag_data_dictionary_spec

        # Add row to table
        table.add_row(variant.short_name, variant.variant_type.value, str(len(all_services)),
                      str(len(ddds.data_object_props)),
                      str(len(getattr(variant, "comparams_refs", []))))
    rich_print(table)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .element import IdentifiableElement
from .exceptions import odxrequire
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, resolve_snref
from .snrefcontext import SnRefContext
from .state import State
from .statetransition import StateTransition
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class StateChart(IdentifiableElement):
    """
    Corresponds to STATE-CHART.
    """
    semantic: str
    state_transitions: list[StateTransition] = field(default_factory=list)
    start_state_snref: str
    states: NamedItemList[State] = field(default_factory=NamedItemList)

    @property
    def start_state(self) -> State:
        return self._start_state

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "StateChart":
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        semantic: str = odxrequire(et_element.findtext("SEMANTIC"))

        state_transitions = [
            StateTransition.from_et(st_elem, context)
            for st_elem in et_element.iterfind("STATE-TRANSITIONS/STATE-TRANSITION")
        ]

        start_state_snref_elem = odxrequire(et_element.find("START-STATE-SNREF"))
        start_state_snref = start_state_snref_elem.attrib["SHORT-NAME"]

        states = [
            State.from_et(st_elem, context) for st_elem in et_element.iterfind("STATES/STATE")
        ]

        return StateChart(
            semantic=semantic,
            state_transitions=state_transitions,
            start_state_snref=start_state_snref,
            states=NamedItemList(states),
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        odxlinks = {self.odx_id: self}

        for strans in self.state_transitions:
            odxlinks.update(strans._build_odxlinks())

        for st in self.states:
            odxlinks.update(st._build_odxlinks())

        return odxlinks

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        for strans in self.state_transitions:
            strans._resolve_odxlinks(odxlinks)

        for st in self.states:
            st._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        context.state_chart = self

        # For now, we assume that the start state short name reference
        # points to a local state of the state chart. TODO: The XSD
        # allows to define state charts without any states, yet the
        # start state SNREF is mandatory. Is this a gap in the spec or
        # does it allow "foreign" start states? If the latter, what
        # does that mean?
        self._start_state = resolve_snref(self.start_state_snref, self.states, State)

        for strans in self.state_transitions:
            strans._resolve_snrefs(context)

        for st in self.states:
            st._resolve_snrefs(context)

        context.state_chart = None



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .dyniddefmodeinfo import DynIdDefModeInfo
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext


@dataclass(kw_only=True)
class DynDefinedSpec:
    dyn_id_def_mode_infos: list[DynIdDefModeInfo] = field(default_factory=list)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DynDefinedSpec":
        dyn_id_def_mode_infos = [
            DynIdDefModeInfo.from_et(x, context)
            for x in et_element.iterfind("DYN-ID-DEF-MODE-INFOS/DYN-ID-DEF-MODE-INFO")
        ]
        return DynDefinedSpec(dyn_id_def_mode_infos=dyn_id_def_mode_infos)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result: dict[OdxLinkId, Any] = {}

        for didmi in self.dyn_id_def_mode_infos:
            result.update(didmi._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        for didmi in self.dyn_id_def_mode_infos:
            didmi._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        for didmi in self.dyn_id_def_mode_infos:
            didmi._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import cast
from xml.etree import ElementTree

from ..exceptions import DecodeError, EncodeError, odxassert, odxraise
from ..odxdoccontext import OdxDocContext
from ..odxtypes import AtomicOdxType, DataType
from ..utils import dataclass_fields_asdict
from .compucategory import CompuCategory
from .compumethod import CompuMethod
from .intervaltype import IntervalType
from .linearsegment import LinearSegment


@dataclass(kw_only=True)
class ScaleLinearCompuMethod(CompuMethod):
    """A piecewise linear compu method which may feature discontinuities.

    For details, refer to ASAM specification MCD-2 D (ODX), section 7.3.6.6.4.
    """

    @property
    def segments(self) -> list[LinearSegment]:
        return self._segments

    @staticmethod
    def compu_method_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                             internal_type: DataType,
                             physical_type: DataType) -> "ScaleLinearCompuMethod":
        cm = CompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
        kwargs = dataclass_fields_asdict(cm)

        return ScaleLinearCompuMethod(**kwargs)

    def __post_init__(self) -> None:
        self._segments: list[LinearSegment] = []

        odxassert(self.category == CompuCategory.SCALE_LINEAR,
                  "ScaleLinearCompuMethod must exibit SCALE-LINEAR category")

        odxassert(self.physical_type in [
            DataType.A_FLOAT32,
            DataType.A_FLOAT64,
            DataType.A_INT32,
            DataType.A_UINT32,
        ])
        odxassert(self.internal_type in [
            DataType.A_FLOAT32,
            DataType.A_FLOAT64,
            DataType.A_INT32,
            DataType.A_UINT32,
        ])

        if self.compu_internal_to_phys is None:
            odxraise("SCALE-LINEAR compu methods require COMPU-INTERNAL-TO-PHYS")
            return

        compu_scales = self.compu_internal_to_phys.compu_scales

        for scale in compu_scales:
            self._segments.append(
                LinearSegment.from_compu_scale(
                    scale, internal_type=self.internal_type, physical_type=self.physical_type))

        # find out if the transfer function is invertible (i.e. if it
        # can be encoded by normal means). section 7.3.6.6.4 of the
        # ODX specification states that the condition for
        # invertibility is that adjacent COMPU-SCALES exhibit the same
        # values on their common boundaries and that the slope in all
        # intervals exhibit the same sign (or are 0). For segments
        # with a slope of zero, COMPU-INVERSE-VALUE shall be used.
        self._is_invertible = True
        ref_factor = self._segments[0].factor
        for i in range(0, len(self._segments) - 1):
            s0 = self.segments[i]
            s1 = self.segments[i + 1]

            if ref_factor * s1.factor < 0:
                self._is_invertible = False
                break
            if s1.factor != 0:
                ref_factor = s1.factor

            # both interval boundaries must not be infinite
            if s0.internal_upper_limit is None or \
               s1.internal_lower_limit is None:
                self._is_invertible = False
                break
            elif s0.internal_upper_limit.value is None or \
               s1.internal_lower_limit.value is None or \
               s0.internal_upper_limit.interval_type == IntervalType.INFINITE or \
               s1.internal_lower_limit.interval_type == IntervalType.INFINITE:
                self._is_invertible = False
                break

            # the intervals must use the same reference point
            if (x := s0.internal_upper_limit.value) != s1.internal_lower_limit.value:
                self._is_invertible = False
                break

            if not isinstance(x, (int, float)):
                odxraise("Linear segments must use int or float for all quantities")

            # the respective function value at the interval's
            # reference point must be identical
            y0 = s0.convert_internal_to_physical(x)
            y1 = s1.convert_internal_to_physical(x)
            if abs(y0 - y1) < 1e-10:
                self._is_invertible = False
                break

    def convert_physical_to_internal(self, physical_value: AtomicOdxType) -> float | int:
        if not self._is_invertible:
            odxraise(
                f"Trying to encode value {physical_value!r} using a non-invertible "
                f"SCALE-LINEAR transfer function", EncodeError)

        applicable_segments = [
            seg for seg in self._segments if seg.physical_applies(physical_value)
        ]
        if not applicable_segments:
            odxraise(r"No applicable segment for value {physical_value} found", EncodeError)
            return cast(int, None)

        seg = applicable_segments[0]

        return seg.convert_physical_to_internal(physical_value)

    def convert_internal_to_physical(self, internal_value: AtomicOdxType) -> float | int:
        applicable_segments = [
            seg for seg in self._segments if seg.internal_applies(internal_value)
        ]
        if not applicable_segments:
            odxraise(r"No applicable segment for value {internal_value} found", DecodeError)
            return cast(int, None)

        seg = applicable_segments[0]
        return seg.convert_internal_to_physical(internal_value)

    def is_valid_physical_value(self, physical_value: AtomicOdxType) -> bool:
        return any(True for seg in self._segments if seg.physical_applies(physical_value))

    def is_valid_internal_value(self, internal_value: AtomicOdxType) -> bool:
        return any(True for seg in self._segments if seg.internal_applies(internal_value))



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from .exceptions import odxraise
from .odxdoccontext import OdxDocContext
from .odxtypes import DataType
from .radix import Radix


@dataclass(kw_only=True)
class PhysicalType:
    """The physical type describes the base data type of a parameter.

    Similar to how a `DiagCodedType` describes the encoding of the internal value,
    the `PhysicalType` describes how to display the physical value.

    For an unsigned integers (A_UINT32) it may specify a display radix (HEX, DEC, BIN, OCT).
    For floating point numbers (A_FLOAT32, A_FLOAT64) it may specify a precision,
    that is, the number of digits to display after the decimal point.

    Examples
    --------

    A hexadecimal, integer number::

    PhysicalType(DataType.A_UINT32, display_radix=Radix.HEX)

    A floating point number that should be displayed with 2 digits after the decimal point::

    PhysicalType(DataType.A_FLOAT64, precision=2)
    """

    precision: int | None = None
    """Number of digits after the decimal point to display to the user
    The precision is only applicable if the base data type is A_FLOAT32 or A_FLOAT64.
    """

    base_data_type: DataType

    display_radix: Radix | None = None
    """The display radix defines how integers are displayed to the user.
    The display radix is only applicable if the base data type is A_UINT32.
    """

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "PhysicalType":
        precision_str = et_element.findtext("PRECISION")
        precision = int(precision_str) if precision_str is not None else None

        base_data_type_str = et_element.get("BASE-DATA-TYPE")
        if base_data_type_str not in DataType.__members__:
            odxraise(f"Encountered unknown base data type '{base_data_type_str}'")
        base_data_type = DataType(base_data_type_str)

        display_radix_str = et_element.get("DISPLAY-RADIX")
        if display_radix_str is not None:
            if display_radix_str not in Radix.__members__:
                odxraise(f"Encountered unknown display radix '{display_radix_str}'")
            display_radix = Radix.__members__[display_radix_str]
        else:
            display_radix = None

        return PhysicalType(
            precision=precision, base_data_type=base_data_type, display_radix=display_radix)



# SPDX-License-Identifier: MIT
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from typing_extensions import override

from .decodestate import DecodeState
from .determinenumberofitems import DetermineNumberOfItems
from .encodestate import EncodeState
from .exceptions import DecodeError, EncodeError, odxassert, odxraise, odxrequire
from .field import Field
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .odxtypes import ParameterValue
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class DynamicLengthField(Field):
    """Array of structure with length field"""
    offset: int
    determine_number_of_items: DetermineNumberOfItems

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DynamicLengthField":
        kwargs = dataclass_fields_asdict(Field.from_et(et_element, context))

        offset = int(odxrequire(et_element.findtext('OFFSET')))
        determine_number_of_items = DetermineNumberOfItems.from_et(
            odxrequire(et_element.find('DETERMINE-NUMBER-OF-ITEMS')),
            context,
        )

        return DynamicLengthField(
            offset=offset, determine_number_of_items=determine_number_of_items, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        odxlinks = super()._build_odxlinks()
        odxlinks.update(self.determine_number_of_items._build_odxlinks())
        return odxlinks

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)
        self.determine_number_of_items._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)
        self.determine_number_of_items._resolve_snrefs(context)

    @override
    def encode_into_pdu(self, physical_value: ParameterValue, encode_state: EncodeState) -> None:

        odxassert(encode_state.cursor_bit_position == 0,
                  "No bit position can be specified for dynamic length fields!")

        if not isinstance(physical_value, Sequence):
            odxraise(
                f"Expected a list of values for dynamic length field {self.short_name}, "
                f"got {type(physical_value)}", EncodeError)

        # move the origin to the cursor position
        orig_origin = encode_state.origin_byte_position
        encode_state.origin_byte_position = encode_state.cursor_byte_position

        det_num_items = self.determine_number_of_items
        encode_state.cursor_bit_position = self.determine_number_of_items.bit_position or 0
        encode_state.cursor_byte_position = encode_state.origin_byte_position + det_num_items.byte_position
        det_num_items.dop.encode_into_pdu(len(physical_value), encode_state)

        if encode_state.cursor_byte_position - encode_state.origin_byte_position > self.offset:
            odxraise(f"The length specifier of field {self.short_name} overlaps "
                     f"with its first item!")

        encode_state.cursor_byte_position = encode_state.origin_byte_position + self.offset
        encode_state.cursor_bit_position = 0

        orig_is_end_of_pdu = encode_state.is_end_of_pdu
        encode_state.is_end_of_pdu = False
        for i, value in enumerate(physical_value):
            if i == len(physical_value) - 1:
                encode_state.is_end_of_pdu = orig_is_end_of_pdu

            self.structure.encode_into_pdu(value, encode_state)
        encode_state.is_end_of_pdu = orig_is_end_of_pdu

        # ensure the correct message size if the field is empty
        if len(physical_value) == 0:
            encode_state.emplace_bytes(b"")

        # move cursor and origin positions
        encode_state.origin_byte_position = orig_origin

    @override
    def decode_from_pdu(self, decode_state: DecodeState) -> ParameterValue:

        odxassert(decode_state.cursor_bit_position == 0,
                  "No bit position can be specified for dynamic length fields!")

        orig_origin = decode_state.origin_byte_position

        det_num_items = self.determine_number_of_items
        decode_state.origin_byte_position = decode_state.cursor_byte_position
        decode_state.cursor_byte_position = decode_state.origin_byte_position + det_num_items.byte_position
        decode_state.cursor_bit_position = det_num_items.bit_position or 0

        n = det_num_items.dop.decode_from_pdu(decode_state)
        result: list[ParameterValue] = []

        if not isinstance(n, int):
            odxraise(f"Number of items specified by a dynamic length field {self.short_name} "
                     f"must be an integer (is: {type(n).__name__})")
        elif n < 0:
            odxraise(
                f"Number of items specified by a dynamic length field {self.short_name} "
                f"must be positive (is: {n})", DecodeError)
            n = 0

        decode_state.cursor_byte_position = decode_state.origin_byte_position + self.offset
        for _ in range(n):
            result.append(self.structure.decode_from_pdu(decode_state))

        decode_state.origin_byte_position = orig_origin

        return result



# SPDX-License-Identifier: MIT
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from typing_extensions import override

from .dataobjectproperty import DataObjectProperty
from .decodestate import DecodeState
from .dynenddopref import DynEndDopRef
from .encodestate import EncodeState
from .exceptions import DecodeError, EncodeError, odxassert, odxraise, odxrequire
from .field import Field
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .odxtypes import AtomicOdxType, ParameterValue
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class DynamicEndmarkerField(Field):
    """Array of a structure with variable length determined by a termination sequence"""

    dyn_end_dop_ref: DynEndDopRef

    @property
    def dyn_end_dop(self) -> DataObjectProperty:
        return self._dyn_end_dop

    @property
    def termination_value(self) -> AtomicOdxType:
        return self._termination_value

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DynamicEndmarkerField":
        kwargs = dataclass_fields_asdict(Field.from_et(et_element, context))

        # ODX 2.0 uses DATA-OBJECT-PROP-REF
        # ODX 2.2 uses DYN-END-DOP-REF
        dop_ref = et_element.find("DYN-END-DOP-REF") or et_element.find("DATA-OBJECT-PROP-REF")
        dyn_end_dop_ref = DynEndDopRef.from_et(odxrequire(dop_ref), context)

        return DynamicEndmarkerField(dyn_end_dop_ref=dyn_end_dop_ref, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        odxlinks = super()._build_odxlinks()
        return odxlinks

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        self._dyn_end_dop = odxlinks.resolve(self.dyn_end_dop_ref, DataObjectProperty)

        tv_string = self.dyn_end_dop_ref.termination_value_raw
        tv_physical = self._dyn_end_dop.diag_coded_type.base_data_type.from_string(tv_string)

        self._termination_value = tv_physical

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

    @override
    def encode_into_pdu(self, physical_value: ParameterValue, encode_state: EncodeState) -> None:

        odxassert(encode_state.cursor_bit_position == 0,
                  "No bit position can be specified for dynamic endmarker fields!")
        if not isinstance(physical_value, Sequence):
            odxraise(
                f"Expected a sequence of values for dynamic endmarker field {self.short_name}, "
                f"got {type(physical_value).__name__}", EncodeError)
            return

        orig_is_end_of_pdu = encode_state.is_end_of_pdu
        encode_state.is_end_of_pdu = False
        for i, item in enumerate(physical_value):
            if i == len(physical_value) - 1:
                encode_state.is_end_of_pdu = orig_is_end_of_pdu

            self.structure.encode_into_pdu(item, encode_state)
        encode_state.is_end_of_pdu = orig_is_end_of_pdu

        if not encode_state.is_end_of_pdu:
            # only add an endmarker if we are not at the end of the
            # PDU. note that since section 7.3.6.10.5 of the MCD-2
            # specification states that the data used by the endmarker
            # ought to be considered to be not consumed (why?!), we
            # need to keep the cursor where it is before adding the
            # endmarker. (we still consider its bits to be used
            # "used", in order to produce a warning if it is attempted
            # to be overridden.)
            tmp_cursor = encode_state.cursor_byte_position
            self.dyn_end_dop.encode_into_pdu(self.termination_value, encode_state)
            encode_state.cursor_byte_position = tmp_cursor

    @override
    def decode_from_pdu(self, decode_state: DecodeState) -> ParameterValue:

        odxassert(decode_state.cursor_bit_position == 0,
                  "No bit position can be specified for dynamic endmarker fields!")

        orig_origin = decode_state.origin_byte_position
        decode_state.origin_byte_position = decode_state.cursor_byte_position

        result: list[ParameterValue] = []
        while True:
            # check if we're at the end of the PDU
            if decode_state.cursor_byte_position == len(decode_state.coded_message):
                break

            # check if the cursor currently points to a termination
            # value
            tmp_cursor = decode_state.cursor_byte_position
            try:
                tv_candidate = self.dyn_end_dop.decode_from_pdu(decode_state)
                if tv_candidate == self.termination_value:
                    # note that section 7.3.6.10.5 of the MCD-2
                    # specification states that the bytes occupied by
                    # the endmarker ought to be considered to be not
                    # consumed (why?!), i.e., we need to keep the
                    # cursor where it is before adding the
                    # endmarker. (we still consider its to be used
                    # "used", though.)
                    decode_state.cursor_byte_position = tmp_cursor
                    break
            except DecodeError:
                pass
            decode_state.cursor_byte_position = tmp_cursor

            result.append(self.structure.decode_from_pdu(decode_state))

        decode_state.origin_byte_position = orig_origin

        return result



#! /usr/bin/python3
#
# SPDX-License-Identifier: MIT
#
# Implementation of the "lazy" somersault ECU
import argparse
import asyncio
import logging
import random
from socket import socket
from typing import cast

import isotp
import isotp.tpsock
import somersaultecu

import odxtools_porsche.uds as uds
from odxtools.exceptions import DecodeError, odxrequire
from odxtools.message import Message
from odxtools.odxtypes import ParameterValueDict
from odxtools.response import Response, ResponseType

from .somersaultecu import database as somersaultdatabase

tester_logger = logging.getLogger("somersault_lazy_tester")
ecu_logger = logging.getLogger("somersault_lazy_ecu")

is_sterile = False
can_channel: str | None = None
somersault_lazy_diag_layer = somersaultdatabase.ecus.somersault_lazy

# the raw payload data of the telegrams received by the ECU and by the
# tester when in sterile mode (unittest without a CAN channel)
sterile_rx_ecu: list[bytes] = []
sterile_rx_ecu_event: asyncio.Event | None = None
sterile_rx_tester: list[bytes] = []
sterile_rx_tester_event: asyncio.Event | None = None


def create_isotp_socket(channel: str | None, rxid: int, txid: int) -> isotp.tpsock.socket | None:
    if is_sterile:
        return None

    assert channel is not None

    # create an ISO-TP socket without a timeout (timeouts are handled
    # using asyncio). Also, for asyncio to work with this socket, it
    # must be non-blocking...
    result_socket = isotp.socket(timeout=None)
    result_socket._socket.setblocking(False)

    # set the ISO-TP flow control options:
    #
    # stmin: minimum frame separation time [ms]
    # bs: maximum block size. (Must be smaller than 4096?)
    result_socket.set_fc_opts(stmin=5, bs=100)

    # instruct the ISO-TP subsystem to pad the last CAN frame of a
    # telegram to 8 bytes of payload using 0xcc padding bytes
    result_socket.set_opts(txpad=0xCC)

    isotp_addr = isotp.Address(rxid=rxid, txid=txid)
    result_socket.bind(channel, address=isotp_addr)

    return result_socket


async def ecu_send(isotp_socket: isotp.tpsock.socket | None, payload: bytes) -> None:
    """
    ECU sends a message, either in "sterile" or in "live" mode.
    """
    global sterile_rx_tester
    global sterile_rx_tester_event

    if is_sterile:
        assert isotp_socket is None
        assert sterile_rx_tester_event is not None

        sterile_rx_tester_event.clear()
        sterile_rx_tester.append(payload)
        sterile_rx_tester_event.set()
    else:
        assert isotp_socket is not None

        loop = asyncio.get_running_loop()
        await loop.sock_sendall(cast(socket, isotp_socket), payload)


async def ecu_recv(isotp_socket: isotp.tpsock.socket | None) -> bytes:
    """
    ECU receives a message, either in "sterile" or in "live" mode.
    """
    global sterile_rx_ecu
    global sterile_rx_ecu_event

    if is_sterile:
        assert isotp_socket is None
        assert sterile_rx_ecu_event is not None

        if len(sterile_rx_ecu) > 0:
            return sterile_rx_ecu.pop(0)

        sterile_rx_ecu_event.clear()
        await sterile_rx_ecu_event.wait()
        sterile_rx_ecu_event.clear()

        return sterile_rx_ecu.pop(0)
    else:
        assert isotp_socket is not None

        loop = asyncio.get_running_loop()
        return await loop.sock_recv(cast(socket, isotp_socket), 4095)


async def tester_send(isotp_socket: isotp.tpsock.socket | None, payload: bytes) -> None:
    """
    Tester sends a message, either in "sterile" or in "live" mode.
    """
    global sterile_rx_ecu
    global sterile_rx_ecu_event

    if is_sterile:
        assert isotp_socket is None
        assert sterile_rx_ecu_event is not None

        sterile_rx_ecu.append(payload)
        sterile_rx_ecu_event.set()
    else:
        assert isotp_socket is not None

        loop = asyncio.get_running_loop()
        await loop.sock_sendall(cast(socket, isotp_socket), payload)


async def tester_recv(isotp_socket: isotp.tpsock.socket | None) -> bytes:
    """
    Tester receives a message, either in "sterile" or in "live" mode.
    """
    global sterile_rx_tester
    global sterile_rx_tester_event

    if is_sterile:
        assert isotp_socket is None
        assert sterile_rx_tester_event is not None

        if len(sterile_rx_tester) > 0:
            return sterile_rx_tester.pop(0)

        sterile_rx_tester_event.clear()
        await sterile_rx_tester_event.wait()
        sterile_rx_tester_event.clear()

        return sterile_rx_tester.pop(0)
    else:
        assert isotp_socket is not None

        loop = asyncio.get_running_loop()
        return await loop.sock_recv(cast(socket, isotp_socket), 4095)


class SomersaultLazyEcu:

    def __init__(self) -> None:
        self._diag_session_open = False
        self._data_receive_event = asyncio.Event()
        self.dizziness_level = 0
        self.max_dizziness_level = 10

        self.isotp_socket = create_isotp_socket(
            can_channel,
            rxid=odxrequire(somersault_lazy_diag_layer.get_can_receive_id()),
            txid=odxrequire(somersault_lazy_diag_layer.get_can_send_id()),
        )

        ##############
        # extract the tester present parameters from the ECU's
        # communication parameters.
        ##############

        # determine the timeout on inactivity [s]
        self._idle_timeout = somersault_lazy_diag_layer.get_tester_present_time() or 3.0

        # we send a response to tester present messages. make sure
        # that this is specified
        cp = odxrequire(somersault_lazy_diag_layer.get_comparam("CP_TesterPresentReqRsp"))
        assert cp.value == "Response expected" or cp.value == "1"

    async def _handle_requests_task(self) -> None:
        while True:
            data = await ecu_recv(self.isotp_socket)

            # to whom it may concern: we have received some data. This
            # is used to make sure that we do not run into the "no
            # tester present" timeout and close the session spuriously
            self._data_receive_event.set()

            ecu_logger.debug(f"Data received: 0x{data.hex()}")

            # decode the data
            try:
                messages = somersault_lazy_diag_layer.decode(data)

                # do the actual work which we were asked to do. we assume
                # that requests can always be uniquely decoded
                assert len(messages) == 1
                await self._handle_request(messages[0])
            except DecodeError as e:
                ecu_logger.warning(f"Could not decode request "
                                   f"0x{data.hex()}: {e}")
                return

    async def _handle_request(self, message: Message) -> None:
        service = message.service

        ecu_logger.info(f"received UDS message: {service.short_name}")

        # keep alive message.
        if service.short_name == "tester_present":
            # send a positive response if have an active diagnostic
            # session, and a negative one if we don't.
            if self._diag_session_open:
                response_payload = service.positive_responses[0].encode(
                    coded_request=message.coded_message)
            else:
                response_payload = service.negative_responses[0].encode(
                    coded_request=message.coded_message)

            await ecu_send(self.isotp_socket, response_payload)
            return

        if service.short_name == "session_start":
            if not self._diag_session_open:
                response_payload = service.positive_responses[0].encode(
                    coded_request=message.coded_message, can_do_backward_flips="false")
            else:
                response_payload = service.negative_responses[0].encode(
                    coded_request=message.coded_message)

            self._diag_session_open = True
            await ecu_send(self.isotp_socket, response_payload)
            return

        # from here on, a diagnostic session must be started or else
        # we will send a generic "ServiceNotSupportedInActiveSession"
        # UDS response
        if not self._diag_session_open:
            rq_id = 0
            if message.coded_message:
                rq_id = message.coded_message[0]

            # send a "service not supported in active session" UDS
            # response.
            await ecu_send(self.isotp_socket, bytes([0x7F, rq_id, 0x7F]))
            return

        # stop the diagnostic session
        if service.short_name == "session_stop":
            if self._diag_session_open:
                response_payload = service.positive_responses[0].encode(
                    coded_request=message.coded_message)
            else:
                response_payload = service.negative_responses[0].encode(
                    coded_request=message.coded_message)

            self._diag_session_open = False
            await ecu_send(self.isotp_socket, response_payload)
            return

        if service.short_name == "do_forward_flips":
            await self._handle_forward_flip_request(message)
            return

    async def _handle_forward_flip_request(self, message: Message) -> None:
        service = message.service
        # TODO: the need for .param_dict is quite ugly IMO,
        # i.e. provide a __getitem__() method for the Message class() (?)
        soberness_check = message.param_dict["forward_soberness_check"]
        num_flips = message.param_dict["num_flips"]

        if soberness_check != 0x12:
            response = service.negative_responses.flips_not_done
            response_data = response.encode(
                coded_request=message.coded_message,
                reason=0,  # -> not sober
                flips_successfully_done=0,
            )
            await ecu_send(self.isotp_socket, response_data)
            return

        # we cannot do all flips because we are too dizzy
        assert isinstance(num_flips, int)
        if self.dizziness_level + num_flips > self.max_dizziness_level:

            response = service.positive_responses.grudging_forward
            response_data = response.encode(
                coded_request=message.coded_message,
                reason=1,  # -> too dizzy
                flips_successfully_done=self.max_dizziness_level - self.dizziness_level,
            )
            await ecu_send(self.isotp_socket, response_data)
            self.dizziness_level = self.max_dizziness_level
            return

        # do the flips, but be aware that 1% of our attempts fail
        # because we stumble
        for i in range(0, num_flips):
            if random.randrange(0, 10000) < 100:
                response = service.negative_responses.flips_not_done
                response_data = response.encode(
                    coded_request=message.coded_message,
                    reason=2,  # -> stumbled
                    flips_successfully_done=i,
                )
                await ecu_send(self.isotp_socket, response_data)
                return

            self.dizziness_level += 1

        response = service.positive_responses.grudging_forward
        response_data = response.encode(coded_request=message.coded_message)
        await ecu_send(self.isotp_socket, response_data)

    def close_diag_session(self) -> None:
        if not self._diag_session_open:
            return

        self._diag_session_open = False

        # clean up data associated with the diagnostic session

    async def run(self) -> None:
        ecu_logger.info("running diagnostic server")

        cst = self._auto_close_session_task()
        hrt = self._handle_requests_task()

        await asyncio.gather(cst, hrt)

    async def _auto_close_session_task(self) -> None:
        # task to close the diagnostic session if the tester has not
        # been seen for longer than the timeout specified by the ECU.
        while True:
            # sleep until we either hit our timeout or we've received
            # some data from the tester
            try:
                await asyncio.wait_for(self._data_receive_event.wait(), self._idle_timeout * 1.05)
            except asyncio.exceptions.TimeoutError:
                # we ran into the idle timeout. Close the diagnostic
                # session if it is open. note that this also happens if
                # there is no active diagnostic session
                self.close_diag_session()
                continue
            except asyncio.exceptions.CancelledError:
                return

            if self._data_receive_event.is_set():
                # we have received some data
                self._data_receive_event.clear()
                continue


async def tester_await_response(isotp_socket: isotp.tpsock.socket | None,
                                raw_message: bytes,
                                timeout: float = 0.5) -> bytes | ParameterValueDict:
    # await the answer from the server (be aware that the maximum
    # length of ISO-TP telegrams over the CAN bus is 4095 bytes)
    raw_response = await tester_recv(isotp_socket)

    tester_logger.debug(f"received response")

    try:
        replies = somersault_lazy_diag_layer.decode_response(raw_response, raw_message)
        assert len(replies) == 1  # replies must always be uniquely decodable
        assert isinstance(replies[0].coding_object, Response)

        if replies[0].coding_object.response_type == ResponseType.POSITIVE:
            rtype = "positive"
        elif replies[0].coding_object.response_type == ResponseType.NEGATIVE:
            rtype = "negative"
        else:
            rtype = "unknown"

        tester_logger.debug(f"received {rtype} response")

        return replies[0].param_dict

    except DecodeError as e:
        if len(raw_response) >= 3:
            sid = raw_response[0]
            rq_sid = raw_response[1]
            error_id = raw_response[2]

            if sid == uds.NegativeResponseId:
                try:
                    rq_name = somersaultecu.SID(rq_sid).name
                except ValueError:
                    rq_name = f"0x{rq_sid:x}"
                error_name = uds.NegativeResponseCodes(error_id).name

                tester_logger.debug(
                    f"Received negative response by service {rq_name}: {error_name}")
                return raw_response

        tester_logger.debug(f"Could not decode response: {e}")
        return raw_response


async def tester_main() -> None:
    tester_logger.info("running diagnostic tester")

    # note that ODX specifies the CAN IDs from the ECU's point of
    # view, i.e., from the tester's (our) perspective, they are
    # reversed.
    isotp_socket = create_isotp_socket(
        can_channel,
        txid=odxrequire(somersault_lazy_diag_layer.get_can_receive_id()),
        rxid=odxrequire(somersault_lazy_diag_layer.get_can_send_id()),
    )

    # try to to do a single forward flip without having an active session (ought to fail)
    tester_logger.debug(f"attempting a sessionless forward flip")
    raw_message = somersault_lazy_diag_layer.services.do_forward_flips(
        forward_soberness_check=0x12, num_flips=1)
    await tester_send(isotp_socket, raw_message)
    await tester_await_response(isotp_socket, raw_message)

    # send "start session"
    tester_logger.debug(f"starting diagnostic session")
    raw_message = somersault_lazy_diag_layer.services.session_start()
    await tester_send(isotp_socket, raw_message)
    await tester_await_response(isotp_socket, raw_message)

    # attempt to do a single forward flip
    tester_logger.debug(f"attempting a forward flip")
    raw_message = somersault_lazy_diag_layer.services.do_forward_flips(
        forward_soberness_check=0x12, num_flips=1)
    await tester_send(isotp_socket, raw_message)
    await tester_await_response(isotp_socket, raw_message)

    # attempt to do a single forward flip but fail the soberness check
    tester_logger.debug(f"attempting a forward flip")
    raw_message = somersault_lazy_diag_layer.services.do_forward_flips(
        forward_soberness_check=0x23, num_flips=1)
    await tester_send(isotp_socket, raw_message)
    await tester_await_response(isotp_socket, raw_message)

    # attempt to do three forward flips
    tester_logger.debug(f"attempting three forward flip")
    raw_message = somersault_lazy_diag_layer.services.do_forward_flips(
        forward_soberness_check=0x12, num_flips=3)
    await tester_send(isotp_socket, raw_message)
    await tester_await_response(isotp_socket, raw_message)

    # attempt to do 50 forward flips (should always fail because of dizzyness)
    tester_logger.debug(f"attempting 50 forward flip")
    raw_message = somersault_lazy_diag_layer.services.do_forward_flips(
        forward_soberness_check=0x12, num_flips=50)
    await tester_send(isotp_socket, raw_message)
    await tester_await_response(isotp_socket, raw_message)

    tester_logger.debug(f"Finished")


async def main(args: argparse.Namespace) -> None:
    global is_sterile
    global sterile_rx_ecu_event
    global sterile_rx_tester_event

    if args.mode == "unittest":
        is_sterile = args.channel is None
        sterile_rx_ecu_event = asyncio.Event()
        sterile_rx_tester_event = asyncio.Event()
    elif args.channel is None:
        print("A CAN channel must be specified when not in unittest mode.")
        return

    # TODO: handle ISO-TP communication parameters (block size,
    # timings, ...) specified by the Somersault ECU
    tester_task = None
    if args.mode != "server":
        tester_task = asyncio.create_task(tester_main())

    server_task = None
    if args.mode != "tester":
        somersault_ecu = SomersaultLazyEcu()
        server_task = asyncio.create_task(somersault_ecu.run())

    if args.mode == "server":
        assert server_task is not None
        await server_task
    elif args.mode == "tester":
        assert tester_task is not None
        await tester_task
    else:
        assert args.mode == "unittest"
        assert server_task is not None
        assert tester_task is not None

        # run both tasks in parallel. Since the server task does not
        # complete, we need to wait until the first task is completed
        # and then manually that this was the server task
        done, pending = await asyncio.wait([tester_task, server_task],
                                           return_when=asyncio.FIRST_COMPLETED)

        if tester_task in pending:
            tester_logger.error("The tester task did not terminate. This "
                                "should *never* happen!")
            tester_task.cancel()

        if server_task not in pending:
            ecu_logger.error("The server task terminated. This "
                             "should *never* happen!")
        else:
            # since if everything goes according to plan only the
            # tester terminates, we need to stop server task here to
            # avoid complaints from asyncio...
            server_task.cancel()


parser = argparse.ArgumentParser(
    description="Provides an implementation for the 'lazy' variant of the somersault ECU")

parser.add_argument(
    "--channel",
    "-c",
    required=False,
    help="CAN interface name to be used (required for tester or server modes)",
)
parser.add_argument(
    "--mode",
    "-m",
    default="unittest",
    required=False,
    help="Specify whether to start the ECU side ('server'), the tester side ('tester') or both ('unittest')",
)

args = parser.parse_args()  # deals with the help message handling

# set the verbosity of the log output
logging.basicConfig(level=logging.DEBUG)  # log messages from the ECU itself
logging.getLogger("odxtools").setLevel(
    logging.WARNING)  # log messages stemming from the odxtools library

can_channel = args.channel

asyncio.run(main(args))



# SPDX-License-Identifier: MIT
import os
from pathlib import Path

from .database import Database


def load_pdx_file(pdx_file: str | Path) -> Database:
    db = Database()
    db.add_pdx_file(str(pdx_file))
    db.refresh()
    return db


def load_odx_d_file(odx_d_file_name: str | Path) -> Database:
    db = Database()
    db.add_odx_file(str(odx_d_file_name))
    db.refresh()

    return db


def load_file(file_name: str | Path) -> Database:
    if str(file_name).lower().endswith(".pdx"):
        return load_pdx_file(str(file_name))
    elif str(file_name).lower().endswith(".odx-d"):
        return load_odx_d_file(str(file_name))
    else:
        raise RuntimeError(f"Could not guess the file format of file '{file_name}'!")


def load_files(*file_names: str | Path) -> Database:
    db = Database()
    for file_name in file_names:
        p = Path(file_name)
        if p.suffix.lower() == ".pdx":
            db.add_pdx_file(str(file_name))
        elif p.suffix.lower().startswith(".odx"):
            db.add_odx_file(str(file_name))
        elif p.name.lower() != "index.xml":
            db.add_auxiliary_file(str(file_name))

    db.refresh()
    return db


def load_directory(dir_name: str | Path) -> Database:
    db = Database()
    for file_name in os.listdir(str(dir_name)):
        p = Path(dir_name) / file_name

        if not p.is_file():
            continue

        if p.suffix.lower() == ".pdx":
            db.add_pdx_file(str(p))
        elif p.suffix.lower().startswith(".odx"):
            db.add_odx_file(str(p))
        elif p.name.lower() != "index.xml":
            db.add_auxiliary_file(p.name, open(str(p), "rb"))

    db.refresh()
    return db



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any, Union
from xml.etree import ElementTree

from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .snrefcontext import SnRefContext
from .specialdata import SpecialData
from .specialdatagroupcaption import SpecialDataGroupCaption


@dataclass(kw_only=True)
class SpecialDataGroup:
    """This corresponds to the SDG XML tag"""
    sdg_caption: SpecialDataGroupCaption | None = None
    sdg_caption_ref: OdxLinkRef | None = None
    values: list[Union["SpecialDataGroup", SpecialData]] = field(default_factory=list)
    semantic_info: str | None = None  # the "SI" attribute

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "SpecialDataGroup":

        sdg_caption = None
        if caption_elem := et_element.find("SDG-CAPTION"):
            sdg_caption = SpecialDataGroupCaption.from_et(caption_elem, context)

        sdg_caption_ref = None
        if (caption_ref_elem := et_element.find("SDG-CAPTION-REF")) is not None:
            sdg_caption_ref = OdxLinkRef.from_et(caption_ref_elem, context)

        semantic_info = et_element.get("SI")

        values: list[SpecialData | SpecialDataGroup] = []
        for value_elem in et_element:
            next_entry: SpecialData | SpecialDataGroup | None = None
            if value_elem.tag == "SDG":
                next_entry = SpecialDataGroup.from_et(value_elem, context)
            elif value_elem.tag == "SD":
                next_entry = SpecialData.from_et(value_elem, context)

            if next_entry is not None:
                values.append(next_entry)

        return SpecialDataGroup(
            sdg_caption=sdg_caption,
            sdg_caption_ref=sdg_caption_ref,
            semantic_info=semantic_info,
            values=values,
        )

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {}

        if self.sdg_caption_ref is None and self.sdg_caption is not None:
            result.update(self.sdg_caption._build_odxlinks())

        for val in self.values:
            result.update(val._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if self.sdg_caption_ref is not None:
            self.sdg_caption = odxlinks.resolve(self.sdg_caption_ref, SpecialDataGroupCaption)
        elif self.sdg_caption is not None:
            # resolve the ODXLINK references of the caption, but only
            # if the caption was specified by value, not by reference
            self.sdg_caption._resolve_odxlinks(odxlinks)

        for val in self.values:
            val._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        # resolve the SNREFs of the caption, but only if the caption
        # was specified by value, not by reference
        if self.sdg_caption is not None and self.sdg_caption_ref is None:
            self.sdg_caption._resolve_snrefs(context)

        for val in self.values:
            val._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import cast
from xml.etree import ElementTree

from ..exceptions import odxassert, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxtypes import DataType


@dataclass(kw_only=True)
class CompuRationalCoeffs:
    value_type: DataType

    numerators: list[int | float] = field(default_factory=list)
    denominators: list[int | float] = field(default_factory=list)

    @staticmethod
    def coeffs_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                       value_type: DataType) -> "CompuRationalCoeffs":
        odxassert(
            value_type
            in (DataType.A_UINT32, DataType.A_INT32, DataType.A_FLOAT32, DataType.A_FLOAT64),
            "Rational coefficients must be of numeric type.")

        numerators = [
            cast(float, value_type.from_string(odxrequire(elem.text)))
            for elem in et_element.iterfind("COMPU-NUMERATOR/V")
        ]
        denominators = [
            cast(float, value_type.from_string(odxrequire(elem.text)))
            for elem in et_element.iterfind("COMPU-DENOMINATOR/V")
        ]

        return CompuRationalCoeffs(
            value_type=value_type, numerators=numerators, denominators=denominators)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .companydata import CompanyData
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .snrefcontext import SnRefContext
from .specialdatagroup import SpecialDataGroup
from .teammember import TeamMember


@dataclass(kw_only=True)
class CompanyDocInfo:
    company_data_ref: OdxLinkRef
    team_member_ref: OdxLinkRef | None = None
    doc_label: str | None = None
    sdgs: list[SpecialDataGroup] = field(default_factory=list)

    @property
    def company_data(self) -> CompanyData:
        return self._company_data

    @property
    def team_member(self) -> TeamMember | None:
        return self._team_member

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "CompanyDocInfo":
        # the company data reference is mandatory
        company_data_ref = odxrequire(
            OdxLinkRef.from_et(et_element.find("COMPANY-DATA-REF"), context))
        team_member_ref = OdxLinkRef.from_et(et_element.find("TEAM-MEMBER-REF"), context)
        doc_label = et_element.findtext("DOC-LABEL")
        sdgs = [SpecialDataGroup.from_et(sdge, context) for sdge in et_element.iterfind("SDGS/SDG")]

        return CompanyDocInfo(
            company_data_ref=company_data_ref,
            team_member_ref=team_member_ref,
            doc_label=doc_label,
            sdgs=sdgs,
        )

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {}

        for sdg in self.sdgs:
            result.update(sdg._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._company_data = odxlinks.resolve(self.company_data_ref, CompanyData)

        self._team_member: TeamMember | None = None
        if self.team_member_ref is not None:
            self._team_member = odxlinks.resolve(self.team_member_ref, TeamMember)

        for sdg in self.sdgs:
            sdg._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        for sdg in self.sdgs:
            sdg._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any, cast
from xml.etree import ElementTree

from .element import IdentifiableElement
from .exceptions import odxraise, odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext
from .standardizationlevel import StandardizationLevel
from .usage import Usage
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class BaseComparam(IdentifiableElement):
    param_class: str
    cptype: StandardizationLevel
    display_level: int | None = None
    cpusage: Usage | None = None  # Required in ODX 2.2, missing in ODX 2.0

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "BaseComparam":
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        param_class = odxrequire(et_element.attrib.get("PARAM-CLASS"))

        cptype_str = odxrequire(et_element.attrib.get("CPTYPE"))
        try:
            cptype = StandardizationLevel(cptype_str)
        except ValueError:
            cptype = cast(StandardizationLevel, None)
            odxraise(f"Encountered unknown CPTYPE '{cptype_str}'")

        dl = et_element.attrib.get("DISPLAY_LEVEL")
        display_level = None if dl is None else int(dl)

        # Required in ODX 2.2, missing in ODX 2.0
        cpusage_str = et_element.attrib.get("CPUSAGE")
        cpusage = None
        if cpusage_str is not None:
            try:
                cpusage = Usage(cpusage_str)
            except ValueError:
                cpusage = None
                odxraise(f"Encountered unknown CPUSAGE '{cpusage_str}'")

        return BaseComparam(
            param_class=param_class,
            cptype=cptype,
            display_level=display_level,
            cpusage=cpusage,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {self.odx_id: self}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        pass

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass



# SPDX-License-Identifier: MIT
from copy import deepcopy
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any
from xml.etree import ElementTree

from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict

if TYPE_CHECKING:
    from .diaglayers.diaglayer import DiagLayer


@dataclass(kw_only=True)
class ParentRef:
    layer_ref: OdxLinkRef
    not_inherited_diag_comms: list[str] = field(default_factory=list)  # short_name references
    not_inherited_variables: list[str] = field(default_factory=list)  # short_name references
    not_inherited_dops: list[str] = field(default_factory=list)  # short_name references
    not_inherited_tables: list[str] = field(default_factory=list)  # short_name references
    not_inherited_global_neg_responses: list[str] = field(
        default_factory=list)  # short_name references

    @property
    def layer(self) -> "DiagLayer":
        return self._layer

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "ParentRef":

        layer_ref = odxrequire(OdxLinkRef.from_et(et_element, context))

        not_inherited_diag_comms = [
            odxrequire(el.get("SHORT-NAME"))
            for el in et_element.iterfind("NOT-INHERITED-DIAG-COMMS/"
                                          "NOT-INHERITED-DIAG-COMM/"
                                          "DIAG-COMM-SNREF")
        ]
        not_inherited_variables = [
            odxrequire(el.get("SHORT-NAME"))
            for el in et_element.iterfind("NOT-INHERITED-VARIABLES/"
                                          "NOT-INHERITED-VARIABLE/"
                                          "DIAG-VARIABLE-SNREF")
        ]
        not_inherited_dops = [
            odxrequire(el.get("SHORT-NAME")) for el in et_element.iterfind("NOT-INHERITED-DOPS/"
                                                                           "NOT-INHERITED-DOP/"
                                                                           "DOP-BASE-SNREF")
        ]
        not_inherited_tables = [
            odxrequire(el.get("SHORT-NAME")) for el in et_element.iterfind("NOT-INHERITED-TABLES/"
                                                                           "NOT-INHERITED-TABLE/"
                                                                           "TABLE-SNREF")
        ]
        not_inherited_global_neg_responses = [
            odxrequire(el.get("SHORT-NAME"))
            for el in et_element.iterfind("NOT-INHERITED-GLOBAL-NEG-RESPONSES/"
                                          "NOT-INHERITED-GLOBAL-NEG-RESPONSE/"
                                          "GLOBAL-NEG-RESPONSE-SNREF")
        ]

        return ParentRef(
            layer_ref=layer_ref,
            not_inherited_diag_comms=not_inherited_diag_comms,
            not_inherited_variables=not_inherited_variables,
            not_inherited_dops=not_inherited_dops,
            not_inherited_tables=not_inherited_tables,
            not_inherited_global_neg_responses=not_inherited_global_neg_responses,
        )

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if TYPE_CHECKING:
            self._layer = odxlinks.resolve(self.layer_ref, DiagLayer)
        else:
            self._layer = odxlinks.resolve(self.layer_ref)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass

    def __deepcopy__(self, memo: dict[int, Any]) -> Any:
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result

        fields = dataclass_fields_asdict(self)
        for name, value in fields.items():
            setattr(result, name, deepcopy(value))

        return result



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any, Union
from xml.etree import ElementTree

from .basecomparam import BaseComparam
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .odxtypes import odxstr_to_bool
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict

ComplexValue = list[Union[str, "ComplexValue"]]


def create_complex_value_from_et(et_element: ElementTree.Element) -> ComplexValue:
    result: ComplexValue = []
    for el in et_element:
        if el.tag == "SIMPLE-VALUE":
            result.append("" if el.text is None else el.text)
        else:
            result.append(create_complex_value_from_et(el))
    return result


@dataclass(kw_only=True)
class ComplexComparam(BaseComparam):
    subparams: NamedItemList[BaseComparam] = field(default_factory=NamedItemList)
    physical_default_value: ComplexValue | None = None
    allow_multiple_values_raw: bool | None = None

    @property
    def allow_multiple_values(self) -> bool:
        return self.allow_multiple_values_raw is True

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "ComplexComparam":
        kwargs = dataclass_fields_asdict(BaseComparam.from_et(et_element, context))

        # to avoid a cyclic import, create_any_comparam_from_et cannot
        # be imported globally. TODO: figure out if this has
        # performance implications
        from .createanycomparam import create_any_comparam_from_et

        # extract the specifications of the sub-parameters and their
        # default values. Due to the quirky way this is defined by the
        # ODX specification, this is a *major* pain in the butt!
        subparams: NamedItemList[BaseComparam] = NamedItemList()
        elems = list(et_element)

        # go to the first COMPARAM or COMPLEX-COMPARAM sub-element
        i = 0
        while i < len(elems):
            if elems[i].tag in ("COMPARAM", "COMPLEX-COMPARAM"):
                break
            i += 1

        # extract the sub-parameters
        while i < len(elems):
            if elems[i].tag not in ("COMPARAM", "COMPLEX-COMPARAM"):
                break

            subparam = create_any_comparam_from_et(elems[i], context)
            subparams.append(subparam)
            i += 1

        # extract the complex physical default value. (what's the
        # purpose of this? the sub-parameters can define their own
        # default values if a default is desired...)
        complex_physical_default_value: ComplexValue | None = None
        if (cpdv_elem := et_element.find("COMPLEX-PHYSICAL-DEFAULT-VALUE")) is not None:
            complex_physical_default_value = create_complex_value_from_et(cpdv_elem)

        allow_multiple_values_raw = odxstr_to_bool(et_element.get("ALLOW-MULTIPLE-VALUES"))

        return ComplexComparam(
            subparams=subparams,
            physical_default_value=complex_physical_default_value,
            allow_multiple_values_raw=allow_multiple_values_raw,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        odxlinks = super()._build_odxlinks()
        for subparam in self.subparams:
            odxlinks.update(subparam._build_odxlinks())
        return odxlinks

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)
        for subparam in self.subparams:
            subparam._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)
        for subparam in self.subparams:
            subparam._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
import argparse
import sys
import traceback
from io import StringIO

from ._parser_utils import SubparsersList


class DummyTool:
    """A tool which acts as a placeholder for a "real" tool that
    could not be loaded for whatever reason.

    The intention is to be able to use tools that are available
    without them being disturbed by the ones which aren't. Tools which
    are not available should subparser appear in the list of available
    tools, but as soon as the functionality is attempted to be used it
    should bail out.
    """

    def __init__(self, tool_name: str, error: Exception):
        self._odxtools_tool_name_ = tool_name
        self._error = error

    def add_subparser(self, subparser_list: SubparsersList) -> None:
        desc = StringIO()

        print(f"Tool '{self._odxtools_tool_name_}' is unavailable: {self._error}", file=desc)
        print(file=desc)
        print(f"Traceback:", file=desc)
        traceback.print_tb(self._error.__traceback__, file=desc)

        subparser_list.add_parser(
            self._odxtools_tool_name_,
            description=desc.getvalue(),
            help="Tool unavailable",
            formatter_class=argparse.RawTextHelpFormatter,
        )

    def run(self, args: argparse.Namespace) -> None:
        print(
            f"Error: Tool '{self._odxtools_tool_name_}' is unavailable: {self._error}",
            file=sys.stderr,
        )
        exit(1)



# SPDX-License-Identifier: MIT
import textwrap
from collections.abc import Iterable
from io import StringIO

from .compumethods.compucodecompumethod import CompuCodeCompuMethod
from .compumethods.identicalcompumethod import IdenticalCompuMethod
from .compumethods.intervaltype import IntervalType
from .compumethods.linearcompumethod import LinearCompuMethod
from .compumethods.linearsegment import LinearSegment
from .compumethods.ratfunccompumethod import RatFuncCompuMethod
from .compumethods.ratfuncsegment import RatFuncSegment
from .compumethods.scalelinearcompumethod import ScaleLinearCompuMethod
from .compumethods.scaleratfunccompumethod import ScaleRatFuncCompuMethod
from .compumethods.texttablecompumethod import TexttableCompuMethod
from .dataobjectproperty import DataObjectProperty
from .dtcdop import DtcDop
from .dynamiclengthfield import DynamicLengthField
from .endofpdufield import EndOfPduField
from .exceptions import odxrequire
from .multiplexer import Multiplexer
from .parameters.codedconstparameter import CodedConstParameter
from .parameters.matchingrequestparameter import MatchingRequestParameter
from .parameters.nrcconstparameter import NrcConstParameter
from .parameters.parameter import Parameter
from .parameters.parameterwithdop import ParameterWithDOP
from .parameters.reservedparameter import ReservedParameter
from .parameters.systemparameter import SystemParameter
from .parameters.tablekeyparameter import TableKeyParameter
from .parameters.tablestructparameter import TableStructParameter
from .paramlengthinfotype import ParamLengthInfoType
from .staticfield import StaticField


def _get_linear_segment_info(segment: LinearSegment) -> str:
    ll = segment.physical_lower_limit
    if ll is None or ll.interval_type == IntervalType.INFINITE:
        ll_str = "(-inf"
    else:
        ll_delim = '(' if ll.interval_type == IntervalType.OPEN else '['
        ll_str = f"{ll_delim}{ll._value!r}"

    ul = segment.physical_upper_limit
    if ul is None or ul.interval_type == IntervalType.INFINITE:
        ul_str = "inf)"
    else:
        ul_delim = ')' if ul.interval_type == IntervalType.OPEN else ']'
        ul_str = f"{ul._value!r}{ul_delim}"

    return f"{ll_str}, {ul_str}"


def _get_rat_func_segment_info(segment: RatFuncSegment) -> str:
    ll = segment.lower_limit
    if ll is None or ll.interval_type == IntervalType.INFINITE:
        ll_str = "(-inf"
    else:
        ll_delim = '(' if ll.interval_type == IntervalType.OPEN else '['
        ll_str = f"{ll_delim}{ll._value!r}"

    ul = segment.upper_limit
    if ul is None or ul.interval_type == IntervalType.INFINITE:
        ul_str = "inf)"
    else:
        ul_delim = ')' if ul.interval_type == IntervalType.OPEN else ']'
        ul_str = f"{ul._value!r}{ul_delim}"

    return f"{ll_str}, {ul_str}"


def parameter_info(param_list: Iterable[Parameter], quoted_names: bool = False) -> str:
    q = "'" if quoted_names else ""
    of = StringIO()
    for param in param_list:
        if isinstance(param, CodedConstParameter):
            of.write(f"{q}{param.short_name}{q}: const = {param.coded_value_raw}\n")
            continue
        elif isinstance(param, MatchingRequestParameter):
            of.write(f"{q}{param.short_name}{q}: <matches request>\n")
            continue
        elif isinstance(param, NrcConstParameter):
            of.write(f"{q}{param.short_name}{q}: const; choices = {param.coded_values}\n")
            continue
        elif isinstance(param, ReservedParameter):
            of.write(f"{q}{param.short_name}{q}: <reserved>\n")
            continue
        elif isinstance(param, SystemParameter):
            of.write(
                f"{q}{param.short_name}{q}: <system; kind = \"{param.sysparam}\">; required = {param.is_required}\n"
            )
            continue
        elif isinstance(param, TableKeyParameter):
            of.write(
                f"{q}{param.short_name}{q}: <optional> table key; table = '{param.table.short_name}'; choices:\n"
            )
            for tr in param.table.table_rows:
                of.write(f"  '{tr.short_name}',\n")

            continue
        elif isinstance(param, TableStructParameter):
            of.write(
                f"{q}{param.short_name}{q}: table struct; key = '{param.table_key.short_name}'; choices:\n"
            )
            for tr in param.table_key.table.table_rows:
                of.write(f"  ('{tr.short_name}',\n")
                of.write(f"   {{\n")
                of.write(
                    textwrap.indent(
                        parameter_info(odxrequire(tr.structure).parameters, True), "    "))
                of.write(f"   }}),\n")

            continue
        elif not isinstance(param, ParameterWithDOP):
            of.write(
                f"{q}{param.short_name}{q}: <unhandled parameter type '{type(param).__name__}'>\n")
            continue

        dop = param.dop
        if dop is None:
            of.write("{q}{param.short_name}{q}: <no DOP>\n")
            continue
        elif isinstance(dop, EndOfPduField):
            of.write(f"{q}{param.short_name}{q}: list({{\n")
            of.write(textwrap.indent(parameter_info(dop.structure.parameters, True), "  "))
            of.write(f"}})\n")
            continue
        elif isinstance(dop, StaticField):
            of.write(f"{q}{param.short_name}{q}: length={dop.fixed_number_of_items}; list({{\n")
            of.write(textwrap.indent(parameter_info(dop.structure.parameters, True), "  "))
            of.write(f"}})\n")
            continue
        elif isinstance(dop, DynamicLengthField):
            of.write(f"{q}{param.short_name}{q}: list({{\n")
            of.write(textwrap.indent(parameter_info(dop.structure.parameters, True), "  "))
            of.write(f"}})\n")
            continue
        elif isinstance(dop, ParamLengthInfoType):
            of.write(f"{q}{param.short_name}{q}: ")
            of.write("<optional> ")
            of.write(f"int; length_key='{dop.length_key.short_name}'\n")
            continue
        elif isinstance(dop, DtcDop):
            of.write(f"{q}{param.short_name}{q}: ")
            of.write(f"DTC; choices:\n")
            for dtc in dop.dtcs:
                if dtc.display_trouble_code is not None:
                    dtc_desc = dtc.text and f"; \"{dtc.text}\""
                    of.write(
                        f"  '{dtc.display_trouble_code}' (0x{dtc.trouble_code:06x}{dtc_desc})\n")
                else:
                    dtc_desc = dtc.text and f" (\"{dtc.text}\")"
                    of.write(f"  0x{dtc.trouble_code:06x}{dtc_desc}\n")
            continue
        elif isinstance(dop, Multiplexer):
            of.write(f"{q}{param.short_name}{q}: ")
            if dop.default_case is not None:
                of.write(f"<optional>")
            of.write(f"multiplexer; choices:\n")
            for mux_case in dop.cases:
                of.write(f"  ({repr(mux_case.short_name)}, {{\n")
                if (struc := mux_case.structure) is not None:
                    of.write(textwrap.indent(parameter_info(struc.parameters, True), "    "))
                of.write(f"   }})\n")
            continue
        elif isinstance(dop, DataObjectProperty):
            # a "simple" DOP
            if (cm := dop.compu_method) is None:
                of.write(f"{q}{param.short_name}{q}: <no compu method>\n")
                continue

            if isinstance(cm, TexttableCompuMethod):
                of.write(f"{q}{param.short_name}{q}: enum; choices:\n")
                for scale in odxrequire(cm.compu_internal_to_phys).compu_scales:
                    val_str = ""
                    if scale.lower_limit is not None:
                        val_str = f"({repr(scale.lower_limit.value)})"

                    if scale.compu_const is None:
                        of.write(f"  <ERROR in ODX data: no value specified>\n")
                    else:
                        vt = scale.compu_const.vt
                        v = scale.compu_const.v
                        if vt is not None:
                            of.write(f"  \"{vt}\" {val_str}\n")
                        else:
                            of.write(f"  {v}\n")

            elif isinstance(cm, IdenticalCompuMethod):
                of.write(f"{q}{param.short_name}{q}: {dop.physical_type.base_data_type}\n")

            elif isinstance(cm, ScaleLinearCompuMethod):
                of.write(f"{q}{param.short_name}{q}: {dop.physical_type.base_data_type}")
                seg_list = [_get_linear_segment_info(x) for x in cm.segments]
                of.write(f"; ranges = {{ {', '.join(seg_list)} }}")

                unit = dop.unit
                unit_str = unit.display_name if unit is not None else None
                if unit_str is not None:
                    of.write(f"; unit: {unit_str}")

                of.write("\n")

            elif isinstance(cm, LinearCompuMethod):
                of.write(f"{q}{param.short_name}{q}: {dop.physical_type.base_data_type}")
                of.write(f"; range: {_get_linear_segment_info(cm.segment)}")

                unit = dop.unit
                unit_str = unit.display_name if unit is not None else None
                if unit_str is not None:
                    of.write(f"; unit: {unit_str}")

                of.write("\n")

            elif isinstance(cm, ScaleRatFuncCompuMethod):
                of.write(f"{q}{param.short_name}{q}: {dop.physical_type.base_data_type}")
                if cm._phys_to_int_segments is None:
                    of.write("<NOT ENCODABLE>")
                else:
                    seg_list = [_get_rat_func_segment_info(x) for x in cm._phys_to_int_segments]
                    of.write(f"; ranges = {{ {', '.join(seg_list)} }}")

                    unit = dop.unit
                    unit_str = unit.display_name if unit is not None else None
                    if unit_str is not None:
                        of.write(f"; unit: {unit_str}")

                    of.write("\n")

            elif isinstance(cm, RatFuncCompuMethod):
                of.write(f"{q}{param.short_name}{q}: {dop.physical_type.base_data_type}")
                if cm._phys_to_int_segment is None:
                    of.write("<NOT ENCODABLE>")
                else:
                    of.write(f"; range: {_get_rat_func_segment_info(cm._phys_to_int_segment)}")

                    unit = dop.unit
                    unit_str = unit.display_name if unit is not None else None
                    if unit_str is not None:
                        of.write(f"; unit: {unit_str}")

                of.write("\n")

            elif isinstance(cm, CompuCodeCompuMethod):
                of.write(f"{q}{param.short_name}{q}: {dop.physical_type.base_data_type}")
                of.write(f"; <programmatic translation>")

                of.write("\n")

            else:
                of.write(
                    f"{q}{param.short_name}{q}: unknown compu method {type(dop.compu_method).__name__}\n"
                )
        else:
            of.write(f"{q}{param.short_name}{q}: <unhandled DOP '{type(dop).__name__}'>\n")

    return of.getvalue()
